/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "MixCastSceneCaptureComponent2D.h"
#include "ExpCamera.h"
#include "Data/SdkData.h"
#include "MixCast.h"

#include "EngineModule.h"
#include "EngineGlobals.h"
#include "RenderingThread.h"
#include "SceneInterface.h"
#include "SceneView.h"
#include "Engine/TextureRenderTarget2D.h"
#include "Runtime/Renderer/Private/PostProcess/SceneRenderTargets.h"
#include "ScreenRendering.h"

void UMixCastSceneCaptureComponent2D::Initialize(bool isForeground)
{
	IsForeground = isForeground;

	bCaptureEveryFrame = false;
	bCaptureOnMovement = false;
	bAlwaysPersistRenderingState = true;
	CaptureSource = ESceneCaptureSource::SCS_FinalColorLDR;
	PrimitiveRenderMode = ESceneCapturePrimitiveRenderMode::PRM_RenderScenePrimitives;

#if UE_4_25_OR_LATER
	bUseRayTracingIfEnabled = true;
#endif

	if (IsForeground)
	{
		ShowFlags.SetFog(false);
#if !UE_4_24_OR_LATER
		ShowFlags.SetAtmosphericFog(false);
#endif

#if !MIXCAST_SHADER_PASSES
		if (SdkData::GetProjectSettings()->ClipMode == EMixCastClipMode::Planar)
			bEnableClipPlane = true;
#endif
	}
}

void UMixCastSceneCaptureComponent2D::BeginPlay()
{
	postOverlayDel.BindUObject(this, &UMixCastSceneCaptureComponent2D::OnPostOverlay_RenderThread);
	postResolveDel.BindUObject(this, &UMixCastSceneCaptureComponent2D::OnPostResolve_RenderThread);

#if MIXCAST_SHADER_PASSES
	postOpaqueDel.BindUObject(this, &UMixCastSceneCaptureComponent2D::OnPostOpaque_RenderThread);

	viewExt = MakeShareable(new FMixCastSceneViewExt);
	viewExt->Parent = this;
	SceneViewExtensions.Add(viewExt);
#endif

#if UE_4_27_OR_LATER
	NeedsSeparateDepthRead = ProjectUtils::GetGraphicsType() == mixcast::data::ExpGraphicsType::DX12;
#else
	NeedsSeparateDepthRead = false;
#endif

	Super::BeginPlay();
}
void UMixCastSceneCaptureComponent2D::EndPlay(const EEndPlayReason::Type EndPlayReason)
{
	Super::EndPlay(EndPlayReason);

#if MIXCAST_SHADER_PASSES
	SceneViewExtensions.Remove(viewExt);
	viewExt.Reset();

	for (auto& sharedTex : sharedOcclTexs)
		sharedTex.Value->BeginDestroy();
	sharedOcclTexs.Reset();

	postOpaqueDel.Unbind();
#endif

	postOverlayDel.Unbind();
	postResolveDel.Unbind();
}

void UMixCastSceneCaptureComponent2D::AssignLayersTarget(UTextureRenderTarget2D* NewColorTarget, UTextureRenderTarget2D* NewAlphaTarget)
{
	TextureTarget = NewColorTarget;
	AlphaTarget = NewAlphaTarget;
}
#if MIXCAST_SHADER_PASSES
void UMixCastSceneCaptureComponent2D::AssignOpaqueTargets(UTextureRenderTarget2D* NewOpaqueTarget, UTextureRenderTarget2D* NewProcessedOpaqueColorTarget)
{
	OpaqueTarget = NewOpaqueTarget;
	ProcessedOpaqueColorTarget = NewProcessedOpaqueColorTarget;
}
#endif

void UMixCastSceneCaptureComponent2D::CaptureFrameDeferred(ExpFrame frame)
{
	RequestedFrame = frame;
	WasFrameRequested = true;
	CaptureSceneDeferred();
}

void UMixCastSceneCaptureComponent2D::UpdateSceneCaptureContents(FSceneInterface* Scene)
{
	if (!WasFrameRequested)
		return;

	CaptureBegin.Broadcast(this);

#if MIXCAST_SHADER_PASSES
	ExpFrame thisFrame = RequestedFrame;

	UTexture2D* occlTex = NULL;
	if (IsForeground) {
		WatchedUnrealSharedTexture* sharedOcclTex = GetSharedOcclusionTex(RequestedFrame.occlusionTex, RequestedFrame.camWidth, RequestedFrame.camHeight);
		sharedOcclTex->EnqueueRead();
		occlTex = sharedOcclTex->GetUnrealTex();
	}
#endif

	ENQUEUE_RENDER_COMMAND(PreMixCastCaptureCommand)(
#if MIXCAST_SHADER_PASSES
		[this, thisFrame, occlTex](FRHICommandListImmediate& /*RHICmdList*/)
		{
			if (viewExt.IsValid())
				viewExt->OcclusionTex = occlTex;
			RenderingFrame = thisFrame;
#else
		[this](FRHICommandListImmediate& /*RHICmdList*/)
		{
#endif
			RegisterRenderEvents();
		}
	);

	Super::UpdateSceneCaptureContents(Scene);	

	ENQUEUE_RENDER_COMMAND(PostMixCastCaptureCommand)(
		[this](FRHICommandListImmediate& /*RHICmdList*/)
		{
			UnregisterRenderEvents();
#if MIXCAST_SHADER_PASSES
			if (viewExt.IsValid())
				viewExt->OcclusionTex = NULL;
#endif
		}
	);

	CaptureEnd.Broadcast(this);

	WasFrameRequested = false;
}
#if MIXCAST_SHADER_PASSES
WatchedUnrealSharedTexture* UMixCastSceneCaptureComponent2D::GetSharedOcclusionTex(void* sharedHandle, unsigned int expectedWidth, unsigned int expectedHeight)
{
	WatchedUnrealSharedTexture** sharedOcclTexCheck = sharedOcclTexs.Find(sharedHandle);
	WatchedUnrealSharedTexture* sharedOcclTex = (sharedOcclTexCheck == NULL) ? NULL : *sharedOcclTexCheck;
	if (sharedOcclTex != NULL)
	{
		if (sharedOcclTex->GetUnrealTex()->GetSurfaceWidth() != expectedWidth || sharedOcclTex->GetUnrealTex()->GetSurfaceHeight() != expectedHeight)
		{
			sharedOcclTex->BeginDestroy();
			sharedOcclTex = NULL;
			sharedOcclTexs.Remove(sharedHandle);
		}
	}
	if (sharedOcclTex == NULL)
	{
		sharedOcclTex = WatchedUnrealSharedTexture::Create(((FMixCast*)&IMixCast::Get())->GetClientProcHandle(), sharedHandle);
		UTexture2D* unrealTex = sharedOcclTex->GetUnrealTex();
		unrealTex->SRGB = 0;
		unrealTex->UpdateResource();
		sharedOcclTexs.Add(sharedHandle, sharedOcclTex);
	}
	return sharedOcclTex;
}
#endif
void UMixCastSceneCaptureComponent2D::RegisterRenderEvents()
{
	auto& renderer = GetRendererModule();

	if (CaptureSource != ESceneCaptureSource::SCS_SceneColorHDR) 
	{
#if UE_4_26_OR_LATER
		postOverlayDelHandle = renderer.RegisterOverlayRenderDelegate(postOverlayDel);
#else
		renderer.RegisterOverlayRenderDelegate(postOverlayDel);
#endif
		postResolveDelHandle = renderer.GetResolvedSceneColorCallbacks().Add(postResolveDel);
	}

#if MIXCAST_SHADER_PASSES
#if UE_4_26_OR_LATER
	postOpaqueDelHandle = renderer.RegisterPostOpaqueRenderDelegate(postOpaqueDel);
#else
	renderer.RegisterPostOpaqueRenderDelegate(postOpaqueDel);
#endif
#endif
}
void UMixCastSceneCaptureComponent2D::UnregisterRenderEvents()
{
	auto& renderer = GetRendererModule();

	if (CaptureSource != ESceneCaptureSource::SCS_SceneColorHDR)
	{
#if UE_4_26_OR_LATER
		renderer.RemoveOverlayRenderDelegate(postOverlayDelHandle);
#else
		renderer.RegisterOverlayRenderDelegate(NULL);
#endif
		renderer.GetResolvedSceneColorCallbacks().Remove(postResolveDelHandle);
	}

#if MIXCAST_SHADER_PASSES
#if UE_4_26_OR_LATER
	renderer.RemovePostOpaqueRenderDelegate(postOpaqueDelHandle);
#else
	renderer.RegisterPostOpaqueRenderDelegate(NULL);
#endif
#endif
}

#if MIXCAST_SHADER_PASSES
void UMixCastSceneCaptureComponent2D::FMixCastSceneViewExt::PostRenderBasePass_RenderThread(FRHICommandListImmediate& RHICmdList, FSceneView& InView)
{
	if (OcclusionTex == NULL ||
#if !UE_4_27_OR_LATER
		OcclusionTex->Resource == NULL ||
#endif
		!OcclusionTex->IsValidLowLevel())
		return;

	float DepthMultiplier = 65.525f * InView.WorldToMetersScale;
	FMixCastShaders::WriteOcclusion(RHICmdList, InView, true, OcclusionTex->Resource->TextureRHI, DepthMultiplier);
}

void UMixCastSceneCaptureComponent2D::OnPostOpaque_RenderThread(FPostOpaqueRenderParameters& Parameters)
{
	if (viewExt->OcclusionTex != NULL &&
#if !UE_4_27_OR_LATER
		viewExt->OcclusionTex->Resource != NULL &&
#endif
		viewExt->OcclusionTex->IsValidLowLevel())
	{
		FSceneView* InView = (FSceneView*)Parameters.Uid;
		float DepthMultiplier = 65.525f * InView->WorldToMetersScale;
		FMixCastShaders::WriteOcclusion(*Parameters.RHICmdList, *InView, false, viewExt->OcclusionTex->Resource->TextureRHI, DepthMultiplier);
	}

	if (RenderingFrame.HasCamFlag(ExpCamFlagBit::SeparateOpaque))
	{
		IRendererModule& r = GetRendererModule();
		if (OpaqueTarget != NULL && OpaqueTarget->IsValidLowLevel())
			FMixCastShaders::ReadOpaque(r, Parameters, *OpaqueTarget, !NeedsSeparateDepthRead);
		
		FMixCastShaders::ClearPostOpaque(*Parameters.RHICmdList, *(FSceneView*)Parameters.Uid);
	}
}
#endif

void UMixCastSceneCaptureComponent2D::OnPostOverlay_RenderThread(FPostOpaqueRenderParameters& Parameters)
{
	CurView = (FSceneView*)Parameters.Uid;
}

void UMixCastSceneCaptureComponent2D::OnPostResolve_RenderThread(FRHICommandListImmediate& RHICmdList, class FSceneRenderTargets& SceneContext)
{
	if (CurView == NULL)
		return;

	if (AlphaTarget != NULL && AlphaTarget->IsValidLowLevel())
		FMixCastShaders::ReadAlpha(GetRendererModule(), RHICmdList, SceneContext, *CurView, *AlphaTarget);

#if MIXCAST_SHADER_PASSES
	if (NeedsSeparateDepthRead && OpaqueTarget != NULL && OpaqueTarget->IsValidLowLevel())
		FMixCastShaders::ReadDepth(GetRendererModule(), RHICmdList, *CurView, *OpaqueTarget);
#endif

	CurView = NULL;
}
