/**********************************************************************************
* Blueprint Reality Inc. CONFIDENTIAL
* 2020 Blueprint Reality Inc.
* All Rights Reserved.
*
* NOTICE:  All information contained herein is, and remains, the property of
* Blueprint Reality Inc. and its suppliers, if any.  The intellectual and
* technical concepts contained herein are proprietary to Blueprint Reality Inc.
* and its suppliers and may be covered by Patents, pending patents, and are
* protected by trade secret or copyright law.
*
* Dissemination of this information or reproduction of this material is strictly
* forbidden unless prior written permission is obtained from Blueprint Reality Inc.
***********************************************************************************/

#include "MixCast.h"
#include "MixCastLog.h"
#include "Capture/ExpCamera.h"
#include "Capture/ExpFrameSender.h"
#include "Data/SdkData.h"
#include "Plugin/PluginAnalytics.h"
#include "Plugin/PluginUpdater.h"
#include "Utility/CommandLineUtil.h"
#include "Utility/MixCastSdkInitializer.hpp"

#include "HAL/IConsoleManager.h"
#include "Interfaces/IPluginManager.h"
#if UE_4_20_OR_LATER
#include "Core/Public/Misc/CoreDelegates.h"
#else
#include "CoreDelegates.h"
#endif

IMPLEMENT_MODULE(FMixCast, MixCast)

FMixCast::FMixCast()
{
	bActive = false;
}

void FMixCast::StartupModule()
{
#if UE_EDITOR
	UPluginProjectSettings::Register();
	FPluginAnalytics::SendReport();
	FPluginUpdater::CheckForUpdates(true);
#endif

	if (FModuleManager::Get().IsModuleLoaded("SteamVR"))
		FModuleManager::Get().LoadModule(FName("MixCast_SteamVR"));
	if (FModuleManager::Get().IsModuleLoaded("OculusHMD"))
		FModuleManager::Get().LoadModule(FName("MixCast_Oculus"));
#if MIXCAST_OPENXR
	if (FModuleManager::Get().IsModuleLoaded("OpenXRHMD"))
		FModuleManager::Get().LoadModule(FName("MixCast_OpenXR"));
#endif

	mixcast::thrift::ThriftUtil::SetUpLogCallback();

	ConfigReadyDelegate = FCoreDelegates::ConfigReadyForUse.AddUObject(SdkData::GetProjectSettings(), &UPluginProjectSettings::ValidateCurrentSettings);
	if (GConfig && GConfig->IsReadyForUse())
		SdkData::GetProjectSettings()->ValidateCurrentSettings();

	sdkInitializer = new MixCastSdkInitializer();
}

void FMixCast::ShutdownModule()
{
	if (IsActive())
		SetActive(false, 0);

	FCoreDelegates::ConfigReadyForUse.Remove(ConfigReadyDelegate);
	ConfigReadyDelegate.Reset();

	delete sdkInitializer;
	sdkInitializer = nullptr;

#if UE_EDITOR
	UPluginProjectSettings::Unregister();
#endif
}

void FMixCast::StartMixCast()
{
	clientConnection = new FServiceCommunicator();
}
void FMixCast::EndMixCast()
{
	SetActive(false, 0);
	delete clientConnection;
	clientConnection = nullptr;
}

void FMixCast::SetActive(bool active, int64_t clientProcId)
{
	if (bActive == active)
		return;

	bActive = active;
	if (bActive) {
		SdkData::GetProjectSettings()->ValidateCurrentSettings();
		clientProcHandle = OpenProcess(PROCESS_DUP_HANDLE, true, clientProcId);
	}
	else if (clientProcHandle) {
		CloseHandle(clientProcHandle);
		clientProcHandle = NULL;

		AExpCamera::DisposeAllCachedQueues();
		FExpFrameSender::DisposeAllCachedQueues();
	}

	if (CommandLineUtil::IncludesSpectatorModeFlag() && GEngine->GameViewport != NULL)
		GEngine->GameViewport->bDisableWorldRendering = bActive;

	ActiveChanged.Broadcast();
}

bool FMixCast::SendExperienceEvent(const FString& EventId)
{
	if (bActive)
	{
		return clientConnection->SendExperienceEvent(EventId);
	}
	return false;
}

FString FMixCast::GetPluginVersion()
{
	TSharedPtr<IPlugin> mixcastPlugin = IPluginManager::Get().FindPlugin(IMixCast::GetPluginName());
	if (!mixcastPlugin.IsValid())
	return "";

	const FPluginDescriptor& Descriptor = mixcastPlugin->GetDescriptor();
	return Descriptor.VersionName;
}
void FMixCast::RegisterSdkActor(AMixCastSdkActor* actor)
{
	EWorldType::Type worldType = actor->GetWorld()->WorldType;
	switch (worldType)
	{
	case EWorldType::Editor:
	case EWorldType::EditorPreview:
	case EWorldType::GamePreview:
	case EWorldType::None:
		return;
	}

	if (activeSdkActors.Contains(actor)) {
		MIXCAST_ERROR("Actor already registered!");
		return;
	}

	if (activeSdkActors.Num() == 0)
	{
		SdkData::GetProjectSettings()->ValidateCurrentSettings();
	}
	else
	{
		MIXCAST_LOG("Multiple MixCast SDK Actors found; MixCast might not work as expected!");
	}

	activeSdkActors.Add(actor);

	if (clientConnection == nullptr)
		StartMixCast();
}
void FMixCast::UnregisterSdkActor(AMixCastSdkActor* actor, EEndPlayReason::Type EndPlayReason)
{
	if (!activeSdkActors.Contains(actor)) {
		MIXCAST_ERROR("Actor not registered!");
		return;
	}

	activeSdkActors.Remove(actor);

	if (clientConnection != nullptr && activeSdkActors.Num() == 0 &&
		(EndPlayReason == EEndPlayReason::EndPlayInEditor || EndPlayReason == EEndPlayReason::Quit))
	{
		EndMixCast();
	}
}
